//  File: c08p109.cpp
//        rename as list.cpp
//  The programmer's guide to C++
//  Copyright (C) 1996,1997 by Adrian P. Robson

#include "list.h"

class List::Node {
public:
   List::Node( int data, List::Node* node = 0 )
      : thedata(data), thenext(node)
      // Initialise  a node with 'data' and connect it
      //    to 'node'.
      // If 'node' not given end of list is indicated.
      {}
   int data()
      // The data value in this node
      {  return thedata;   }
   List::Node* next()
      // The next node in the list
      //    or zero if there is not one.
      {  return thenext;   }
   List::Node* splice( List::Node* node )
      // Put 'node' into the list after this node.
      //   'node' is connected to the node this node
      //      was connected to.
      {  node->thenext = thenext;
         thenext = node;
         return node;   }
private:
   int thedata;
   List::Node* thenext;
};

List::List()
    : theroot(0), thecurrent(0), theprevious(0) {}

void List::before( int data )
{
   if ( attop_empty() )
      // insert at 'theroot'
      theroot = thecurrent = new Node(data,theroot);
   else
      // insert after 'theprevious'
      thecurrent = theprevious->splice(new Node(data));
}

void List::after( int data )
{
   if ( empty() )
      // attach to 'theroot'
      thecurrent = theroot = new Node(data);
   else
      if ( end() ) {
         // insert after the 'theprevious'
         thecurrent = theprevious->splice(new Node(data));
      }
      else {
         // insert after 'thecurrent'
         theprevious = thecurrent;
         thecurrent = thecurrent->splice(new Node(data));
      }
}

void List::next()
{
   if ( !end() ) {
      theprevious = thecurrent;
      thecurrent = thecurrent->next();
   }
}

bool List::empty()
{
   return theroot == 0;
}

bool List::end()
{
   return thecurrent == 0;
}

bool List::attop_empty()
{
   return thecurrent == theroot;
}

bool List::atbottom()
{
   return thecurrent->next() == 0;
}

void List::top()
{
   theprevious = 0;
   thecurrent = theroot;
}

void List::bottom()
{
   top();
   if ( !empty() )
      // step down list
      while ( !atbottom() )
         next();
}

int List::current()
{
   return thecurrent->data();
}

void List::clear()
{
   Node* dead;
   while ( theroot != 0 ) {
      // remove and delete top node
      dead = theroot;
      theroot = theroot->next();
      delete dead;
   }
   // set to initial state
   thecurrent = theprevious = 0;
}

List::~List()
{
   clear();
}

