//  File: c09p129.cpp
//  The programmer's guide to C++
//  Copyright (C) 1996,1997 by Adrian P. Robson

#include <iostream.h>
#include <iomanip.h>
#include <math.h>

class Money {
// addition operators
   friend Money operator+( const Money& a, const Money& b );
   friend Money operator+( long double a, const Money& b );
   friend Money operator+( const Money& a, long double b );
// subtraction operators
   friend Money operator-( const Money& a, const Money& b );
   friend Money operator-( long double a, const Money& b );
   friend Money operator-( const Money& a, long double b );
// multiplication operators
   friend Money operator*( const Money& a, const Money& b );
   friend Money operator*( long double a, const Money& b );
   friend Money operator*( const Money& a, long double b );
// division operators
   friend Money operator/( const Money& a, const Money& b );
   friend Money operator/( long double a, const Money& b );
   friend Money operator/( const Money& a, long double b );
// input output operators
   friend ostream& operator<<( ostream& os, const Money& om );
   friend istream& operator>>( istream& is, Money& im );
public:
   Money();
      // Initalise to zero
   Money( long double avalue );
      // Initalise to rounded 'avalue'
   operator long double() const;
      // conversion to arithmetic type
   static long double round();
      // Value of roundings accumulator
   static void clear_round();
      // Reset rounding acumulator to zero
   static void hold_round();
      // Stop accumulating rounding errors
   static void release_round();
      // Restart accumulating rounding errors
private:
   long double value;           // money value
   static long double rounds;   // rounding accumulator
   static bool updating;        // accumulation flag
   void doupdate( long double newvalue );
      // update stored value and rounding accumulator
};

long double Money::rounds = 0;
bool Money::updating = true;

Money operator+( const Money& a, const Money& b )
{
   Money result;
   result.doupdate( a.value + b.value );
   return result;
}

Money operator+( long double a, const Money& b )
{
   Money result;
   result.doupdate( a + b.value );
   return result;
}

Money operator+( const Money& a, long double b )
{  
   Money result;
   result.doupdate( a.value + b );
   return result; 
}

Money operator-( const Money& a, const Money& b )
{  
   Money result;
   result.doupdate( a.value - b.value );
   return result; 
}

Money operator-( long double a, const Money& b )
{  
   Money result;
   result.doupdate( a - b.value );
   return result; 
}

Money operator-( const Money& a, long double b )
{  
   Money result;
   result.doupdate( a.value - b );
   return result; 
}

Money operator*( const Money& a, const Money& b )
{  
   Money result;
   result.doupdate( a.value * b.value );
   return result; 
}

Money operator*( long double a, const Money& b )
{  
   Money result;
   result.doupdate( a * b.value );
   return result; 
}

Money operator*( const Money& a, long double b )
{  
   Money result;
   result.doupdate( a.value * b );
   return result; 
}

Money operator/( const Money& a, const Money& b )
{
   Money result;
   result.doupdate( a.value / b.value );
   return result;
}

Money operator/( long double a, const Money& b )
{
   Money result;
   result.doupdate( a / b.value );
   return result;
}

Money operator/( const Money& a, long double b )
{
   Money result;
   result.doupdate( a.value / b );
   return result;
}

ostream& operator<<( ostream& os, const Money& om )
{
   os << om.value;
   return os; 
}

istream& operator>>( istream& is, Money& im )
{  
   long double input;
   is >> input;
   im.doupdate(input);
   return is;
}

Money::Money() : value(0.0) {}

Money::Money( long double avalue )
{
   doupdate(avalue);
}

Money::operator long double() const
{  
   return value; 
}

long double Money::round()
{
   return rounds; 
}

void Money::clear_round()
{
   rounds = 0;
}

void Money::hold_round()
{
   updating = false;
}

void Money::release_round()
{
   updating = true;
}

void Money::doupdate( long double newvalue )
{
   value = floor(( newvalue + 0.005 ) * 100) / 100;
   if ( updating )
      rounds += newvalue - value;
}


void main()
{
   Money pay = 30;
   cout << "roundings " << Money::round() << endl;
   Money cash = 30.6;
   Money tax(3);
   Money cost = 1.005;
   cout << cost << endl;
   cout << "roundings " << Money::round() << endl;
   cost = tax + 44.445;
   cout << cost << endl;
   cout << "roundings " << Money::round() << endl;
   cost = tax + tax;
   cout << cost << endl;
   cost = 44 + tax;
   cout << cost << endl;
   cout << Money::round() << endl;
   cout << "round hold test" << endl;
   Money::clear_round();
   cout << Money::round() << endl;
   Money::hold_round();
   pay = 123.005;
   cout << Money::round() << endl;
   Money::release_round();
   pay = 123.005;
   cout << pay << endl;
   cout << Money::round() << endl;
}
