//  File: c12p168.cpp
//  The programmer's guide to C++
//  Copyright (C) 1996,1997 by Adrian P. Robson

#include <iostream.h>

template< class Type >
class Array {
public:
	Array( int low, int high );
		// subscript in range low to high
		// initialise to default constructor of Type
	void put( int loc, const Type& value );
		// store value in element at loc.
	Type get( int loc ) const;
		// recover value from element at loc.
	bool inbound( int loc ) const ;
		// check loc is within bounds for this array
	~Array();
private:
	Type* store;
	int lowbound;
	int highbound;
   Array( Array& ) {}
   Array& operator=( Array& ) { return *this; }
};

template< class Type >
Array<Type>::Array( int low, int high )
	: lowbound(low), highbound(high)
{
	store = new Type[highbound - lowbound + 1];
}

template< class Type >
void Array<Type>::put( int loc, const Type& value )
{
	store[loc - lowbound] = value;
}

template< class Type >
Type Array<Type>::get( int loc ) const
{
	return store[loc - lowbound];
}

template< class Type >
bool Array<Type>::inbound( int loc ) const
{
	if ( loc >= lowbound && loc <= highbound )
		return true;
	else
		return false;
}

template< class Type >
Array<Type>::~Array()
{
	delete[] store;
}

//************************
class Probe {
// A trace class with standard copy semantics.
public:
	Probe()
		{ cout << "Probe default constructor" << endl; }
	Probe( const Probe& )
		{ cout << "Probe copy constructor" << endl; }
	Probe& operator=( const Probe& p)
		{ cout << "Probe assignment" << endl;
		  return *this; }
	~Probe()
		{ cout << "Probe destructor" << endl; }
};

void probe()
{
   cout << "->starting probe" << endl;
   cout << "->defining instances of Probe" << endl;
   Probe p1;
   cout << "-> defining the array" << endl;
   Array<Probe> parray(1,3);
   cout << "->putting to the array" << endl;
   parray.put(2,p1);
   cout << "->ending probe" << endl;
}
//************************


void main()
{
   Array<int> a1(5,10);
   if ( a1.inbound(5) )
      cout << "okay" << endl;
   else
      cout << "WOOPS" << endl;
   if ( a1.inbound(10) )
      cout << "okay" << endl;
   else
      cout << "WOOPS" << endl;
   if ( !a1.inbound(4) )
      cout << "okay" << endl;
   else
      cout << "WOOPS" << endl;
   if ( !a1.inbound(11) )
      cout << "okay" << endl;
   else
      cout << "WOOPS" << endl;
   a1.put(3,99);
   cout << "value at 3 is " << a1.get(3) << endl;

   Array<int> a2(4,5);
// /* These are not allowed...
   a2 = a1;             // prevented
   Array<int> a3 = a2;  // prevented
//*/
   probe();

}
