//  File: c13p179.cpp
//  The programmer's guide to C++
//  Copyright (C) 1996,1997 by Adrian P. Robson

#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

class Exception {
public:
   Exception( const char message[] )
      {  strncpy(themessage,message,MAXMESS);
         themessage[MAXMESS-1] = 0;  }
   virtual const char* what()
      {  return themessage;   }
   virtual ~Exception() {}
protected:
   enum { MAXMESS = 100 };
private:
   char themessage[MAXMESS];
};

class Xstock : public Exception {
public:
   Xstock()
      : Exception("Non-specific stock error") {}
protected:
   Xstock( const char message[] )
      : Exception(message) {}
};

class Xinvalbin : public Xstock {
public:
   Xinvalbin( const char abin[] )
      : Xstock("Invalid bin name")
      {  strncpy(thebin,abin,MAXBIN);
         thebin[MAXBIN-1] = 0;  }
   const char* bin()
      {  return thebin; }
private:
   enum { MAXBIN = 32 };
   char thebin[MAXBIN];
};

class Xunlist : public Xstock {
public:
   Xunlist( int anitem )
      : Xstock("Unlisted item"), theitem(anitem)
      {  char temp[20];
         strcpy(fullmess,Xstock::what());
         strcat(fullmess," ");
         strcat(fullmess,itoa(theitem,temp,10));  }
   const char* what()
      {  return fullmess;  }
   int item() const
      {  return theitem;   }
private:
   int theitem;
   char fullmess[Exception::MAXMESS + 10];
};

class StockItem {
public:
   void action()
      { // this is some stock processing that 
        // goes wrong and throws an exception
        throw Xstock();
        // or one of these might be thrown instead
        // throw Xunlist(1);
        // throw Xinvalbin("blue widget"); }
};


void main()
{
   try{
      StockItem test;
      test.action();  // action that may go wrong
   }
   catch ( Xinvalbin xcptn ){
      cout << xcptn.what()
           << " error in bin called " << xcptn.bin() << endl;
   }
   catch ( Xstock& xcptn ){
      cout << xcptn.what() << endl;
   }
   catch ( ... ){
      cout << "unexpected exception" << endl;
   }
}
